/*
	$Id: PathElement.java 5949 2012-02-08 14:35:29Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.path;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.taglib.StylableTag;

/**
	This tag defines an element of a path. The element is the
	part that is visible to the user as a clickable link to
	some other page.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;p:pathelement
      id=....
      style=...
      title=...
      href=...
      tooltip=...
      visible=true|false
      enabled=true|false
   &gt;
</pre>

	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>The ID of the path element. The value of this attribute
		goes directly into the standard HTML <code>id</code> attribute.
		This means that if you need a reference to the menuitem object in your own
		JavaScript code, you may for example use the following code:
		<p>
		<code>
		pathElement = document.getElementById('&lt;id&gt;');
		</code>
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different menu items.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>yes</td>
		<td>The text that should be displayed in the menu for this menuitem.</td>
	</tr>
	<tr>
		<td>href</td>
		<td>-</td>
		<td>no</td>
		<td>
			The URL that this item links to, or null if it shouldn't be linked.
		</td>
	</tr>
	<tr>
		<td>tooltip</td>
		<td>-</td>
		<td>no</td>
		<td>An optional tooltip that is shown as a popup text while the user
		is pointing the mouse over the menuitem. Not supported by all browsers.</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>Set to false to hide the path element completely.</td>
	</tr>
	<tr>
		<td>enabled</td>
		<td>true</td>
		<td>no</td>
		<td>Set to false to not link the path. This is an alternative to set the URL to null.</td>
	</tr>
	</table>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2012-02-08 15:35:29 +0100 (Wed, 08 Feb 2012) $
*/
public class PathElement
	extends StylableTag
{
	/**
		
	*/
	private static final long serialVersionUID = 3580594490341605165L;

	/**
		The parent Path.
	*/
	private Path path = null;

	/**
		The title of the menu item.
	*/
	private String title = null;
	
	/**
		URL to the linked page.
	*/
	private String href = null;
	
	/**
		 A tooltip explaining the link.
	*/
	private String tooltip = null;
	
	private boolean visible = true;
	private boolean enabled = true;
	
	public PathElement()
	{
		super("pathelement");
	}
	
	/*
		Taglib initialization methods
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}
	
	public void setHref(String href)
	{
		this.href = href;
	}
	public String getHref()
	{
		return href;
	}
	
	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}
	
	/**
		@since 2.10
	*/
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	/**
		@since 2.10
	*/
	public boolean isVisible()
	{
		return visible;
	}
	
	/**
		@since 2.10
	*/
	public void setEnabled(boolean enabled)
	{
		this.enabled = enabled;
	}
	/**
		@since 2.10
	*/
	public boolean isEnabled()
	{
		return enabled;
	}
	
	public int doStartTag() 
		throws JspException
	{
		if (!isVisible()) return SKIP_BODY;
		if (!(getParent() instanceof Path)) throw new JspTagException("Tag <p:pathelement> must be inside a <p:path> tag");
		path = (Path)getParent();

		StringBuilder sb = new StringBuilder();
		sb.append(path.getSeparator());
		sb.append("<span");
		addIdAndStyles(sb);
		sb.append(">");
		if (getHref() != null && isEnabled())
		{
			sb.append("<a href=\"").append(getHref()).append("\"");
			if (getTooltip() != null) sb.append(" title=\"").append(getTooltip()).append("\"");
			sb.append(">").append(getTitle()).append("</a>");
		}
		else
		{
			sb.append(getTitle());
		}
		sb.append("</span>");
		try
		{
			pageContext.getOut().print(sb.toString());
    	}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}
}


