/*
	$Id: Table.java 5960 2012-02-15 10:00:03Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.clients.web.taglib.Page;
import net.sf.basedb.clients.web.taglib.StylableTag;
import net.sf.basedb.clients.web.util.HTML;
import net.sf.basedb.util.formatter.Formatter;

import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Arrays;
import java.util.Set;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.http.HttpServletRequest;

/**
	This is the main tag for the table. It must be located inside
	a {@link Page &lt;base:page&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:table
      id=...
      clazz=....
      style=...
      title=...
      sortby=...
      direction=ASC|DESC
      columns=...
      sc=...
      item=...
      subcontext=...
      action=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>
			The ID of the table. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
table = document.getElementById('&lt;id&gt;');
</pre>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			An optional title that will be displayed in <code>&lt;h3&gt;</code>
			format directly above the table.
		</td>
	</tr>
	<tr>
		<td>sortby</td>
		<td>-</td>
		<td>no</td>
		<td>
			Matches the <code>property</code> attribute on a <code>&lt;tbl:columdef&gt; tag.
			This column will then be marked with an upwards or downward pointer
			according to the <code>direction</code> attribute.
		</td>
	</tr>
	<tr>
		<td>direction</td>
		<td>asc</td>
		<td>no</td>
		<td>
			The value should be either <code>asc</code> or <code>desc</code> to
			indicate if the rows are sorted in ascending or descending order. 
			Other values are interpreted as <code>asc</code>.
		</td>
	</tr>
	<tr>
		<td>columns</td>
		<td>-</td>
		<td>no</td>
		<td>
			A comma-separatd list of column ID:s. Only columns with an ID in the list
			will be shown in the table. The special value <code>all</code> shows all
			columns. Individual columns may override this setting by
			specifying <code>always</code> or <code>never</code> instead of the default 
			<code>auto</code> value in their <code>show</code> attribute.
		</td>
	</tr>
	<tr>
		<td>action</td>
		<td>name of current page</td>
		<td>no</td>
		<td>
			The action property of the &lt;form&gt; tag.
		</td>
	</tr>
	<tr>
		<td>sc</td>
		<td>-</td>
		<td>no</td>
		<td>
			The current SessionControl object. Used by some child taglibs for 
			generating context sensitive information.
		</td>
	</tr>
	<tr>
		<td>item</td>
		<td>-</td>
		<td>no</td>
		<td>
			The current Item object. Used by some child taglibs for 
			generating context sensitive information (ie. PresetSelector).
		</td>
	</tr>
	<tr>
		<td>subcontext</td>
		<td>"" (empty string)</td>
		<td>no</td>
		<td>
			The name of the current subcontext. Used by some child taglibs for 
			generating context sensitive information (ie. PresetSelector).
		</td>
	</tr>
	</table>
	<p>
	The <code>&lt;tbl:table&gt;</code> may contain the following subtags:
	<ul>
	<li>{@link Toolbar &lt;tbl:toolbar&gt;}: Creates a toolbar with buttons
	<li>{@link Navigator &lt;tbl:navigator&gt;}: Creates the page navigator
	<li>{@link Data &lt;tbl:data&gt;}: Begins the data section
	</ul>
	
	@author Nicklas
	@version 2.0
*/
public class Table
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = -6717117667917148241L;

	public Table()
	{
		super("itemlist");
	}
	
	/**
		The parent &lt;base:page&gt; tag.
	*/
	private Page page;

	/**
		An optional title of the table.
	*/
	private String title = null;

	/**
		The column to sort by. Should match the sort attribute on the column tag.
	*/
	private String sortby = null;
	private Set<String> sortbyMultiple = null;
	
	private String action = null;
	
	private transient SessionControl sc = null;
	
	private Item itemType = null;
	
	private String subcontext = "";
	
	/**
		ASC or DESC
	*/
	private ItemContext.SortDirection sortDirection = ItemContext.SortDirection.ASC;

	/**
		Which columns that should be visible and their order
	*/
	private String columns = "all";
	
	
	private StringBuilder hiddenForm;
	
	private StringBuilder script;
	
	/**
		Keep track of which class to use for the next row.
	*/
	private int evenodd;
	
	/**
		The class names for even and odd row respectively.
	*/
	private static final String[] rowclass = new String[] { "evenrow", "oddrow" } ;

	/**
		"all" was specified for the columns attribute.
	*/
	private boolean allVisible;

	/**
		The ID of all visible columns, sorted by the order they should appear.
	*/
	private Set<String> visibleColumns;
	
	/**
		The ID of all defined columns
	*/
	private Set<String> definedColumns;

	private Map<String, String> columnContent;
	private Map<String, String> columnFilter;
	private transient Map<String, Formatter<?>> columnFormatter;
	private List<String> hiddenFilteredColumns;
	
	private int numHeaderColumns;
	private int numFilteredColumns;

	/*
		Taglib initialization methods
	*/
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	public void setColumns(String columns)
	{
		this.columns = columns;
	}
	public String getColumns()
	{
		return columns;
	}

	public void setSortby(String sortby)
	{
		this.sortby = sortby;
		if (sortby != null) sortbyMultiple = new HashSet<String>(Arrays.asList(sortby.split(",")));
	}
	public String getSortby()
	{
		return sortby;
	}
	
	public void setAction(String action)
	{
		this.action = action;
	}
	public String getAction()
	{
		return action;
	}
	
	public void setSc(SessionControl sc)
	{
		this.sc = sc;
	}
	public SessionControl getSc()
	{
		return sc;
	}
	
	public void setItem(Item itemType)
	{
		this.itemType = itemType;
	}
	public Item getItem()
	{
		return itemType;
	}
	
	public void setSubcontext(String subcontext)
	{
		this.subcontext = subcontext;
	}
	public String getSubcontext()
	{
		return subcontext;
	}


	public void setDirection(String direction)
	{
		direction = direction.toUpperCase();
		this.sortDirection = ItemContext.SortDirection.valueOf(direction);
	}
	public void setDirection(ItemContext.SortDirection direction)
	{
		this.sortDirection = direction;
	}
	public ItemContext.SortDirection getDirection()
	{
		return sortDirection == null ? ItemContext.SortDirection.ASC : sortDirection;
	}

	/*
		Package methods
	*/
	Page getPage()
	{
		return page;
	}
	void addScript(String theScript)
	{
		script.append(theScript);
	}
	void addColumnDef(ColumnDef cd)
	{
		String columnId = cd.getId();
		boolean alwaysShow = "always".equals(cd.getShow());
		boolean alwaysHide = "never".equals(cd.getShow());
		if (alwaysHide)
		{
			visibleColumns.remove(columnId);
		}
		else if (alwaysShow || allVisible)
		{
			visibleColumns.add(columnId);
		}
		definedColumns.add(columnId);
		if (cd.getFormatter() != null) columnFormatter.put(columnId, cd.getFormatter());
		
		script.append("Table.addColumn('").append(getId()).append("', new Column('").append(cd.getId()).append("'");
		script.append(",'").append(HTML.javaScriptEncode(cd.getTitle())).append("'");
		script.append(cd.getProperty() == null ? ",null" : ",'"+HTML.javaScriptEncode(cd.getProperty())+"'");
		script.append(cd.getSortproperty() == null ? ",null" : ",'"+HTML.javaScriptEncode(cd.getSortproperty())+"'");
		script.append(cd.getFilterproperty() == null ? ",null" : ",'"+HTML.javaScriptEncode(cd.getFilterproperty())+"'");
		script.append(cd.getExportproperty() == null ? ",null" : ",'"+HTML.javaScriptEncode(cd.getExportproperty())+"'");
		script.append(cd.getDatatype() == null ? ",null" : ",'"+cd.getDatatype()+"'");
		script.append(cd.isCollection() ? ",true" : ",false");
		script.append(cd.isAnnotation() ? ",true" : ",false");
		script.append(isColumnVisible(columnId) ? ",true" : ",false");
		script.append(alwaysShow ? ",true" : ",false");
		script.append(alwaysHide ? ",true" : ",false");
		script.append(cd.isSortable() ? ",true" : ",false");
		script.append(cd.isFilterable() ? ",true" : ",false");
		script.append(cd.isExportable() ? ",true" : ",false");
		script.append("));\n");
	}
	
	void endColumnDefinitions()
	{
		visibleColumns.retainAll(definedColumns);
	}
	
	boolean isColumnVisible(String columnId)
	{
		return allVisible || visibleColumns.contains(columnId);
	}
	Set<String> getVisibleColumns()
	{
		return visibleColumns;
	}
	Formatter<?> getFormatter(String columnId)
	{
		return columnFormatter.get(columnId);
	}
	void setColumnContent(String columnId, String content)
	{
		columnContent.put(columnId, content);
	}
	Map<String, String> getColumnContent()
	{
		return columnContent;
	}
	void setColumnFilter(String columnId, String filterHtml)
	{
		columnFilter.put(columnId, filterHtml);
	}
	void addFilteredColumn(String columnId, boolean visible)
	{
		numFilteredColumns++;
		if (!visible) hiddenFilteredColumns.add(columnId);
	}
	List<String> getHiddenFilteredColumns()
	{
		return hiddenFilteredColumns;
	}
	int writeColumns()
		throws Exception
	{
		for (String column : visibleColumns)
		{
			String content = columnContent.get(column);
			if (content == null) content = "<td class=\"cell\"></td>";
			pageContext.getOut().print(content);
		}
		columnContent.clear();
		return visibleColumns.size();
	}
	int writeColumnFilters()
		throws Exception
	{
		for (String column : visibleColumns)
		{
			String content = columnFilter.get(column);
			if (content == null) content = "<th class=\"propertyfilter\"></td>";
			pageContext.getOut().print(content);
		}
		columnFilter.clear();
		return visibleColumns.size();
	}
	boolean isSortedby(String sortcolumn)
	{
		return (sortbyMultiple != null) && sortbyMultiple.contains(sortcolumn);
	}
	ItemContext.SortDirection getReverseDirection()
	{
		return sortDirection == ItemContext.SortDirection.ASC ? ItemContext.SortDirection.DESC : ItemContext.SortDirection.ASC;
	}
	String getDirectionImage()
	{
		return "<img src=\""+getPage().getRoot()+"images/" + 
			(sortDirection == ItemContext.SortDirection.ASC ? "sort_asc.png" : "sort_desc.png") +
			"\" alt=\"\">";
	}

	String getNextRowClass()
	{
		evenodd = 1-evenodd;
		return rowclass[evenodd];
	}
	
	void addHidden(String name, String value)
	{
		hiddenForm.append("<input type=\"hidden\" name=\"").append(name).append("\" value=\"");
		if (value != null) hiddenForm.append(value);
		hiddenForm.append("\">\n");
	}
	
	int numVisibleColumns()
	{
		return visibleColumns.size();
	}
	int getNumFilteredColumns()
	{
		return numFilteredColumns;
	}
	
	void setNumColumns(int numColumns)
	{
		if (numColumns > numHeaderColumns)
		{
			numHeaderColumns += numColumns;
		}
	}
	int numHeaderColumns()
	{
		return numHeaderColumns;
	}
	
	/*
		Display methods
	*/
	public int doStartTag()
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <tbl:table> must be inside a <base:page> tag");

		// Always start at row 0
		evenodd = 0;
		numHeaderColumns = 0;
		numFilteredColumns = 0;
		columnContent = new HashMap<String, String>();
		columnFilter = new HashMap<String, String>();
		columnFormatter = new HashMap<String, Formatter<?>>();
		hiddenFilteredColumns = new LinkedList<String>();
		if ((columns == null) || columns.equals("all"))
		{
			allVisible = true;
			visibleColumns = new LinkedHashSet<String>();
		}
		else
		{
			allVisible = false;
			visibleColumns = new LinkedHashSet<String>( Arrays.asList(columns.split(",") ));
			visibleColumns.remove("-"); // "-" represents the required columns
		}
		definedColumns = new HashSet<String>();

		hiddenForm = new StringBuilder();
		script = new StringBuilder();
		addHidden("ID", pageContext.getRequest().getParameter("ID"));
		addHidden("cmd", "UpdateContext");
		addHidden("columns", getColumns());
		addHidden("sortby", getSortby());
		addHidden("direction", getDirection().name());
		
		if (sc != null && itemType != null)
		{
			ItemContext cc = sc.getCurrentContext(itemType, subcontext == null ? "" : subcontext);
			Object exclude = cc.getObject("exclude");
			if (exclude instanceof String)
			{
				addHidden("exclude", (String)exclude);
			}
		}
		
		// Submit to same page by default
		String action = getAction();
		if (action == null) action = ((HttpServletRequest)pageContext.getRequest()).getRequestURI();
		StringBuilder sb = new StringBuilder();
		sb.append("<div ");
		addIdAndStyles(sb);
		sb.append("><form ");
		if (getId() != null) sb.append(" name=\"").append(getId()).append("\"");
		if (action != null) sb.append(" action=\"").append(action).append("\"");
		sb.append(" method=\"post\" onsubmit=\"return false;\">\n");
		
		sb.append(hiddenForm);
		hiddenForm = new StringBuilder();
		
		if (getTitle() != null) sb.append("<h3 class=\"docked\">").append(getTitle()).append("</h3>\n");
		//sb.append("<table>\n");

		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_BODY_INCLUDE;
	}

	public int doEndTag()
		throws JspException
	{
		StringBuilder sb = new StringBuilder();
		//sb.append("</table>\n");
		sb.append(hiddenForm);
		sb.append("</form>");
		sb.append("</div>");
		sb.append("\n<script language=\"JavaScript\">\n");
		sb.append("Main.onLoad(function() {\n");
		sb.append(script);
		sb.append("});\n");
		sb.append("\n</script>\n");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return EVAL_PAGE;
	}

}

