/**
	$Id: BioWell.java 5566 2011-02-09 13:49:51Z nicklas $

	Copyright (C) 2008 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.MeasuredBioMaterialData;
import net.sf.basedb.core.data.PlateCoordinate;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.Coordinate;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
	A biowell is located on a {@link BioPlate} and contains a {@link MeasuredBioMaterial}.
	The item holds information about witch row and column it is located on the bio plate.

	@author Martin
	@version 2.10
	@see BioPlate
	@base.modified $Date: 2008-11-27 16:33:55 +0100 (Thu, 27 Nov 2008) $
*/
public class BioWell 
	extends ChildItem<BioWellData>
{
	/**
		The type of item represented by this class.
		@see Item#BIOWELL
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOWELL;
		
	/**
		Get a <code>BioWell</code> item when you know the id.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>BioWell</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static BioWell getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioWell bw = dc.loadItem(BioWell.class, id);
		if (bw == null) throw new ItemNotFoundException("BioWell[id="+id+"]");
		return bw;
	}

	
	/**
		Get a query configured to retrieve wells for a plate. A plate is
		always required, but if a null bioplate parameter is used the query
		will include a parameter "bioPlate" that must be set before the
		query is actuall executed. The gain is that the same query can be
		re-used with multiple plates.
		
		@param bioPlate The plate to retreive wells for, null is allowed
			but a bioplate must be set for the "bioPlate" query parameter
			before the query is executed
		@return An {@link ItemQuery} object
		@see Plate#getEvents()
	*/
	public static ItemQuery<BioWell> getQuery(BioPlate bioPlate)
		throws InvalidDataException
	{
		ItemQuery<BioWell> query = new ItemQuery<BioWell>(BioWell.class, null);
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property("bioPlate"),
					Hql.entityParameter("bioPlate", Item.BIOPLATE)
				)
			);
		
		if (bioPlate != null) query.setPermanentEntityParameter("bioPlate", bioPlate);
		return query;
	}
	
	BioWell(BioWellData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
	 	Adds the {@link MeasuredBioMaterial},used in this well, to the Set.
	 */
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		MeasuredBioMaterial bioMaterial = getBioMaterial();
		if (bioMaterial != null) using.add(new ItemProxy(bioMaterial.getId(), bioMaterial.getType()));
		return using;
	}
	
	/**
		Check if a {@link MeasuredBioMaterial} is linked to this bio-well
		@return TRUE if this biowell has a biomaterial connected., FALSE otherwise.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		return getData().getBioMaterial() != null || super.isUsed();
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	@Override
	SharedData getSharedParent()
	{
		return getData().getBioPlate();
	}
	// -------------------------------------------
	/**
		Get the {@link BioPlate} this biowell belongs to.
		@return The <code>BioPlate</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the bioplate
		@throws BaseException If there is another error
	*/
	public BioPlate getPlate()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioPlate.class, getData().getBioPlate());
	}
	
	/**
	 	Gets the {@link MeasuredBioMaterial} connected to this biowell.
	  	@return a <code>MeasuredBioMaterial</code> item or null.
	  	@throws PermissionDeniedException If the logged in user doesn't have 
	  		{@link Permission#READ} permission for the biomaterial.
	  	@throws BaseException If there is another error
	 */
	public MeasuredBioMaterial getBioMaterial()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(MeasuredBioMaterial.class, getData().getBioMaterial());
	}
	
	/**
	 	Gets the row index where this well is located.
	  	@return Row number as an <code>int</code>
	 */
	public int getRow()
	{
		return getData().getRow();
	}
	
	/**
	 	Gets the column index where this well is located.
	  	@return Column number as an <code>int</code>
	 */
	public int getColumn()
	{
		return getData().getColumn();
	}
	
	/**
		Get the plate coordinate representation of the coordinate for this well.
		@return A PlateCoordinate object
		@since 2.17
	*/
	public PlateCoordinate getPlateCoordinate()
	{
		return new PlateCoordinate(getRow(), getColumn());
	}
	
	/**
		Get a string representation of the coordinate for this well. Rows are mapped to
		letters (0-A, 1-B, etc.). Columns are mapped to 1-based numbers (0-1, 1-2, etc.)
		@since 2.16
	*/
	public String getCoordinate()
	{
		return Coordinate.numericToAlpha(getRow()+1) + (getColumn()+1);
	}
	
	/**
		Is this well empty?
		@since 2.16
	*/
	public boolean isEmpty()
	{
		return getData().getBioMaterial() == null;
	}
	
	/**
		Get the lock mode of this well. The lock mode determines if it is possible
		to add biomaterial to an empty well or remove biomaterial from the well.
		@return The lock mode as determined by the plate's plate type
		@since 2.16
	*/
	public LockMode getLockMode()
	{
		return BioWell.LockMode.fromValue(getData().getBioPlate().getBioPlateType().getLockMode());
	}
	
	/**
		Check if it is possible to add/change biomaterial in this well.
		
		@return TRUE if it is possible to add biomaterial to the well, FALSE otherwise
		@since 2.16
		@see LockMode#canAdd(BioWell)
	*/
	public boolean canAddBioMaterial()
	{
		return getLockMode().canAdd(getData());
	}
	
	/**
		Check if it is possible to clear the current biomaterial in this well.
		
		@return TRUE if it is possible to clear the well
		@since 2.16
		@see LockMode#canClear(BioWell)
	*/
	public boolean canClearBioMaterial()
	{
		return getLockMode().canClear(getData());
	}
	
	/**
		If the original biomaterial in this well has been moved to
		another plate (eg. for final storage in a freezer). This property
		is only defined for well with lock mode={@link LockMode#LOCKED_AFTER_MOVE}.
		@return TRUE if the original biomaterial has been moved
		@since 2.16
	*/
	public boolean hasBeenUsed()
	{
		return getData().getBioMaterial() == null && getData().getOriginalBioMaterial() != null;
	}
	
	/**
		Get the original biomaterial that was placed in this biowell. This
		property is only defined for well with lock mode={@link LockMode#LOCKED_AFTER_MOVE}
		after the biomaterial has been moved.
		
		@return The original biomaterial
		@since 2.16
	*/
	public MeasuredBioMaterial getOriginalBioMaterial()
	{
		return getDbControl().getItem(MeasuredBioMaterial.class, getData().getOriginalBioMaterial());
	}
	
	/**
		The lock mode for a well is determined by the plate's bio
		plate type. See {@link BioPlateType#getLockMode()}. The lock-mode
		is used to determine if the well can be assigned to a biomaterial
		or not.
	
		@since 2.16
	*/
	public enum LockMode
	{
		/**
			The well is not locked. It is possible to change the biomaterial
			any number of times.
		*/
		UNLOCKED(0, "Unlocked")
		{
			/**
				@return Always true
			*/
			@Override
			boolean canClear(BioWellData well)
			{
				return true;
			}
			/**
				@return Always true
			*/
			@Override
			boolean canAdd(BioWellData well)
			{
				return true;
			}
		},
		
		/**
			It is allowed to add biomaterial once to the well. It can not be replaced
			by a new biomaterial but it is possible to clear it. The original biomaterial
			is then saved in the {@link BioWell#getOriginalBioMaterial()}.
		*/
		LOCKED_AFTER_MOVE(1, "Locked after add+clear")
		{
			/**
				@return Always true
			*/
			@Override
			boolean canClear(BioWellData well)
			{
				return true;
			}
			/**
				Can only add if the well is empty and hasn't been used already.
			*/
			@Override
			boolean canAdd(BioWellData well)
			{
				return well.getBioMaterial() == null && well.getOriginalBioMaterial() == null;
			}
			/**
				This lockmode will save the original biomaterial in 
				{@link BioWell#getOriginalBioMaterial()} when it is cleared.
				After that, it is not possible to add biomaterial to the well again.
			*/
			@Override
			void onClear(BioWellData well, MeasuredBioMaterialData biomaterial)
			{
				well.setOriginalBioMaterial(biomaterial);
			}
		},
		
		/**
			It is allowed to add biomaterial once to the well. It can't be replaced
			by a new biomaterial and it is not possible to clear it. It is possible to
			add biomaterial to empty wells at any time.
		*/
		LOCKED_AFTER_ADD(2, "Locked after add") 
		{
			/**
				@return TRUE if empty, FALSE otherwise
			*/
			@Override
			boolean canClear(BioWellData well)
			{
				return well.getBioMaterial() == null;
			}
			/**
				Can only add if the well is empty.
			*/
			@Override
			boolean canAdd(BioWellData well)
			{
				return well.getBioMaterial() == null;
			}
		},

		/**
			It is allowed to add biomaterial once to the well, but it most be done
			at the same time the plate is creted. Once the transaction has been
			committed to the database, all wells become locked.
		*/
		LOCKED_AFTER_CREATE(3, "Locked at plate creation") 
		{
			/**
				@return TRUE if empty, FALSE otherwise
			*/
			@Override
			boolean canClear(BioWellData well)
			{
				return well.getBioMaterial() == null;
			}
			/**
				Can only add to unsaved items.
			*/
			@Override
			boolean canAdd(BioWellData well)
			{
				return well.getId() == 0;
			}
		};		
		
		/**
			Maps an integer to a lock mode.
		*/
		private static final Map<Integer, LockMode> valueMapping = new HashMap<Integer, LockMode>();
		
		static
		{
			for (LockMode lockMode : LockMode.values())
			{
				LockMode lm = valueMapping.put(lockMode.getValue(), lockMode);
				assert lm == null : "Another lock mode with the value "+lockMode.getValue()+" already exists";
			}
		}
		
		/**
			The integer value of this lock mode. 
		*/
		private final int value;
		private final String displayValue;
		
		private LockMode(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		@Override
		public String toString()
		{
			return displayValue;
		}

		/**
			Get the integer value that is used when storing a lock mode to the database.
			@return The integer value for this lock mode
		*/
		public int getValue()
		{
			return value;
		}

		/**
			Get the <code>Lock mode</code> object when you know the integer code.
			
			@param value The integer value
			@return The location for the integer value
		*/
		public static LockMode fromValue(int value)
		{
			LockMode lockMode = valueMapping.get(value);
			assert lockMode != null : "lockMode == null for value "+value;
			return lockMode;
		}

		/**
			Is it allowed to clear a well from biomaterial? This method should return
			true for an empty well.
			@param well The biowell to check
		*/
		public boolean canClear(BioWell well)
		{
			return canClear(well.getData());
		}
		
		abstract boolean canClear(BioWellData well);
		
		/**
			Called when a well is cleared from an existing biomaterial. The
			default implementation of this method does nothing, but it may be
			overriden by some lock modes.
			
			@param well The well that is cleared
			@param biomaterial The biomaterial that is currently located on the well
		*/
		void onClear(BioWellData well, MeasuredBioMaterialData biomaterial)
		{}

		/**
			Is it allowed to add (a different) biomaterial to the well? 
			If this method is called for a non-empty well, the check must
			assume that the well is first cleared.
			
			@param well A biowell
			@see BioWell#canAddBioMaterial()
		*/
		public boolean canAdd(BioWell well)
		{
			return canAdd(well.getData());
		}
		
		abstract boolean canAdd(BioWellData well);

		/**
			Called when biomaterial is added to a well. The
			default implementation of this method does nothing, but it may be
			overriden by some lock modes.
			
			@param well The well the biomaterial is added to
			@param biomaterial The biomaterial that is going to be added to the well
		*/
		void onAdd(BioWellData well, MeasuredBioMaterialData biomaterial)
		{}
		
	}
	
}