/*
	$Id: DataCube.java 5818 2011-10-21 11:15:41Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.sql.SQLException;

import net.sf.basedb.core.data.DataCubeData;
import net.sf.basedb.core.data.SharedData;
import net.sf.basedb.core.query.Aggregations;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlResult;
import net.sf.basedb.util.jep.JepFunction;

/**
	A data cube is the main unit for an experiment to store analysed data.
	A location in the cube can be described by three coordinates:
	<code>position</code>, <code>column</code> and <code>layer</code>.
	
	<p>
	<b>The position coordinate</b><be>
	The position is similar to the position of raw data in a raw bioassay.
	The position coordinate can be linked to a single reporter. 
	This means that all analysed data with 
	that position coordinate is related to the specified reporter. Normally, 
	this means that all raw data spots should have the same reporter, but this
	is not checked by the data cube. There may be exceptions, since a merge
	operation may merge raw data spots which have different, but still related, 
	reporters (for example different reporters for the same gene). The position 
	coordinate can then be linked to a a generic reporter for that gene.
	
	<p>
	<b>The column coordinate</b>
	The column corresponds to a single bioassay which can have links
	to any number of raw bioassays. The combination of position and column 
	coordinate may be linked to any number of raw data spots.
	
	This means that all analysed data with that column/position coordinate has 
	been calculated from the specified raw data spots. It is not checked that
	the specified raw data spots actually are members of the raw bioassays
	that the bioassay is linked to.
	
	<p>
	<b>The layer coordinate</b>
	The layer corresponds more or less to a bioassayset. All analysed data
	with a given layer is part of the same bioassayset. However, it is possible
	that more than one bioassayset use the same layer. For example, a filtering
	bioassayset that doesn't change the data uses the same layer as it's parent
	bioassayset.
	
	<img src="doc-files/datacube.png">
	
	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-10-21 13:15:41 +0200 (Fri, 21 Oct 2011) $
	@see <a href="../../../../../html/developer/api/dynamic_api.html">Developer documentation: The Dynamic API</a>
*/
public class DataCube
	extends ChildItem<DataCubeData>
{
	/**
		The type of item represented by this class.
		@see Item#DATACUBE
		@see #getType()
	*/
	public static final Item TYPE = Item.DATACUBE;

	/**
		Create a new <code>DataCube</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param virtualDb The virtual database the cube is located in
		@return The new <code>DataCube</code> item
		@throws BaseException If there is an error
	*/
	static DataCube getNew(DbControl dc, VirtualDb virtualDb)
		throws BaseException
	{
		DataCube cube = dc.newItem(DataCube.class);
		cube.setVirtualDb(virtualDb);
		cube.setCubeNo(virtualDb.getNextCubeNo());
		return cube;
	}

	/**
		Get a <code>VirtualDb</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>VirtualDb</code> item
		@throws ItemNotFoundException If an item with the specified id 
			is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static DataCube getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		DataCube cube = dc.loadItem(DataCube.class, id);
		if (cube == null) throw new ItemNotFoundException("DataCube[id=" + id + "]");
		return cube;
	}

	private long addedBytes = 0;

	/**
	 	Creates a new data cube item from the given data.
		@param data the data
	*/
	DataCube(DataCubeData data)
	{
		super(data);
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From BasicItem
		-------------------------------------------
	*/
	/**
		Delete data and update disk usage information on the 
		experiment if action == Action.DELETE
	*/
	void onBeforeCommit(Transactional.Action action)
		throws BaseException
	{
		super.onBeforeCommit(action);
		if (action == Transactional.Action.DELETE)
		{
			getExperiment().addBytes(-getBytes());
			getVirtualDb().deleteAllCubeData(getCubeNo());
		}
		else if (action == Transactional.Action.CREATE)
		{
			getData().setBytes(addedBytes);
			getData().setMaxRawMappingsForSpot(countSpotMappings());
			addedBytes = 0;
		}
		else if (action == Transactional.Action.UPDATE && addedBytes != 0)
		{
			/*
				In case of an update we handle the bytes ourselves since plugins with 
				long running times has a high risk of getting ItemModifiedException
				if some other thing has happened to the experiment at the same time.
			*/
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = 
				HibernateUtil.getPredefinedQuery(session, "UPDATE_BYTES_FOR_DATACUBE");
			/*
				UPDATE DataCubeData dcd
				SET dcd.bytes = dcd.bytes + :addedBytes
				WHERE dcd.id = :dataCube
			*/
			query.setLong("addedBytes", addedBytes);
			query.setInteger("dataCube", getId());
			query.executeUpdate();
			addedBytes = 0;
		}
	}
	// -------------------------------------------
	/*
		From the ChildItem class
		-------------------------------------------
	*/
	/**
		Get the experiment.
	*/
	SharedData getSharedParent()
	{
		return getData().getVirtualDb().getExperiment();
	}
	/**
		USE permission is enough to be able to manage data cubes.
	*/
	Permission getPermissionForWriteDeleteAndCreate()
	{
		return Permission.USE;
	}
	// -------------------------------------------

	/**
		Get the raw data type this experiment uses for the raw data.
		@return A <code>RawDataType</code> object
	*/
	public RawDataType getRawDataType()
	{
		return RawDataTypes.getSafeRawDataType(getData().getVirtualDb().getExperiment().getRawDataType());
	}
	
	/**
		Get the experiment this data cube belongs to.
		@return An <code>Experiment</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the experiment
		@throws BaseException If there is another error
	*/
	public Experiment getExperiment()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Experiment.class, getData().getVirtualDb().getExperiment());
	}
	
	/**
		Get the virtual database this cube is a part of.
		@return A <code>VirtualDb</code> object
		@throws PermissionDeniedException If the logged in user doesn't have 
			read permission to the item
		@throws BaseException If there is another error
	*/
	VirtualDb getVirtualDb()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(VirtualDb.class, getData().getVirtualDb());
	}
	
	/**
		Set the experiment. This cannot be changed after the dynamic tables 
		has been created.
	*/
	private void setVirtualDb(VirtualDb virtualDb)
		throws PermissionDeniedException, InvalidDataException
	{
		if (virtualDb == null) throw new InvalidUseOfNullException("virtualDb");
		virtualDb.checkPermission(Permission.WRITE);
		getData().setVirtualDb(virtualDb.getData());
	}

	/**
		Get the number of this data cube.
	*/
	public short getCubeNo()
	{
		return getData().getCubeNo();
	}
	private void setCubeNo(short cubeNo)
	{
		getData().setCubeNo(cubeNo);
	}
	
	/**
		Get the number of filters created for this data cube. 
	*/
	public short getNumFilters()
	{
		return getData().getNumFilters();
	}
	
	/**
		Generate the next available filter number. This method will open a
		second transaction and update the {@link DataCubeData#getNumFilters()}
		property by one, unless this item hasn't been saved to the database yet.
		@return The next filter number
	*/
	short getNextFilterNo()
		throws BaseException
	{
		short filterNo = 0;
		if (!isInDatabase())
		{
			// Just increase counter
			filterNo = (short)(getNumFilters() + 1);
		}
		else
		{
			// Update the db
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Increase the number of filter
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_FILTERNO_FOR_DATACUBE");
				/*
					UPDATE DataCubeData dcb
					SET dcb.numFilters = dcb.numFilters + 1
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				HibernateUtil.executeUpdate(query);
				
				// Get the new value
				query = HibernateUtil.getPredefinedQuery(session, "GET_FILTERNO_FOR_DATACUBE");
				/*
					SELECT dcb.numFilters
					FROM DataCubeData dcb
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				filterNo = HibernateUtil.loadData(Short.class, query);
				HibernateUtil.commit(tx);
				
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}

		// Update data object and return
		getData().setNumFilters(filterNo);
		return filterNo;
	}

	
	/**
		Get the number of extra values created for this data cube. 
	*/
	public short getNumExtraValues()
	{
		return getData().getNumExtraValues();
	}
	
	/**
		Generate the next available extra value number. This method will open a
		second transaction and update the {@link DataCubeData#getNumExtraValues()}
		property by one, unless this item hasn't been saved to the database yet.
		@return The next extra value number
	*/
	short getNextExtraValueNo()
		throws BaseException
	{
		short extraNo = 0;
		if (!isInDatabase())
		{
			// Just increase counter
			extraNo = (short)(getNumExtraValues() + 1);
		}
		else
		{
			// Update the db
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Increase the number of extra values
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_EXTRAVALUENO_FOR_DATACUBE");
				/*
					UPDATE DataCubeData dcb
					SET dcb.numExtraValues = dcb.numExtraValues + 1
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				HibernateUtil.executeUpdate(query);
				
				// Get the new value
				query = HibernateUtil.getPredefinedQuery(session, "GET_EXTRAVALUENO_FOR_DATACUBE");
				/*
					SELECT dcb.numExtraValues
					FROM DataCubeData dcb
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				extraNo = HibernateUtil.loadData(Short.class, query);
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
		// Update data object and return
		getData().setNumExtraValues(extraNo);
		return extraNo;
	}
	
	/**
		Get the number of layers created in this data cube. 
	*/
	public short getNumLayers()
	{
		return getData().getNumLayers();
	}
	
	/**
		Generate the next available layer number. This method will open a
		second transaction and update the {@link DataCubeData#getNumLayers()}
		property by one, unless this item hasn't been saved to the database yet.
		@return The next layer number
	*/
	short getNextLayerNo()
		throws BaseException
	{
		short layerNo = 0;
		if (!isInDatabase())
		{
			// Just increase counter
			layerNo = (short)(getNumLayers() + 1);
		}
		else
		{
			// Update the db
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Increase the number of extra values
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_LAYERNO_FOR_DATACUBE");
				/*
					UPDATE DataCubeData dcb
					SET dcb.numLayers = dcb.numLayers + 1
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				HibernateUtil.executeUpdate(query);
				
				// Get the new value
				query = HibernateUtil.getPredefinedQuery(session, "GET_LAYERNO_FOR_DATACUBE");
				/*
					SELECT dcb.numLayers
					FROM DataCubeData dcb
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				layerNo = HibernateUtil.loadData(Short.class, query);
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}

		// Update data object and return
		getData().setNumLayers(layerNo);
		return layerNo;
	}
	
	/**
		Get the number of columns created in this data cube. 
	*/
	public short getNumColumns()
	{
		return getData().getNumColumns();
	}
	
	/**
		Generate the next available column number. This method will open a
		second transaction and update the {@link DataCubeData#getNumColumns()}
		property by one, unless this item hasn't been saved to the database yet.
		@return The next column number
	*/
	short getNextColumnNo()
		throws BaseException
	{
		short columnNo = 0;
		if (!isInDatabase())
		{
			// Just increase counter
			columnNo = (short)(getNumColumns() + 1);
		}
		else
		{
			// Update the db
			org.hibernate.Session session = null;
			org.hibernate.Transaction tx = null;
			try
			{
				session = HibernateUtil.newSession();
				tx = HibernateUtil.newTransaction(session);
				
				// Increase the number of extra values
				org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, "UPDATE_COLUMNNO_FOR_DATACUBE");
				/*
					UPDATE DataCubeData dcb
					SET dcb.numColumns = dcb.numColumns + 1
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				HibernateUtil.executeUpdate(query);
				
				// Get the new value
				query = HibernateUtil.getPredefinedQuery(session, "GET_COLUMNNO_FOR_DATACUBE");
				/*
					SELECT dcb.numColumns
					FROM DataCubeData dcb
					WHERE dcb.id = :dataCube
				*/
				query.setInteger("dataCube", this.getId());
				columnNo = HibernateUtil.loadData(Short.class, query);
				HibernateUtil.commit(tx);
			}
			catch (BaseException ex)
			{
				if (tx != null) HibernateUtil.rollback(tx);
				throw ex;
			}
			finally
			{
				if (session != null) HibernateUtil.close(session);
			}
		}
	
		// Update data object and return
		getData().setNumColumns(columnNo);
		return columnNo;
	}
	
	/**
		Get the number of positions created in this data cube. 
	*/
	public int getNumPositions()
	{
		return getData().getNumPositions();
	}
	/**
		Set by the {@link PositionBatcher#close()} method. Cannot be
		updated after that.
	*/
	void setNumPositions(int positions)
	{
		getData().setNumPositions(positions);
	}
	
	/**
		Get the number of mappings to raw data spots for the spot with
		the most number of mappings. For most data cubes this value is 
		probably 1. After a merge, the count can be higher which may require some
		plugins to behave differently when using values from the raw data table.
		@since 2.2
	*/
	public int getMaxRawMappingsForSpot()
	{
		return getData().getMaxRawMappingsForSpot();
	}
	
	/**
	  	Gets the number of bytes the analysed data in this cube occupies on disk.
	*/
	public long getBytes()
	{
		return getData().getBytes();
	}
	void addBytes(long bytes)
	{
		addedBytes += bytes;
		getExperiment().addBytes(bytes);
	}

	/**
		Create a new layer in this datacube.
	*/
	DataCubeLayer newLayer()
	{
		return DataCubeLayer.getNew(getDbControl(), this);
	}
	
	/**
		Create a new column in this database.
	*/
	DataCubeColumn newColumn()
	{
		return DataCubeColumn.getNew(getDbControl(), this);
	}
	
	/**
		Create a new filter in this datacube.
	*/
	DataCubeFilter newFilter()
	{
		return DataCubeFilter.getNew(getDbControl(), this);
	}
	
	/**
		Create a new extra vlaue in this data cube.
	*/
	DataCubeExtraValue newExtraValue(Type valueType, ExtraValue.CoordinateType coordinateType)
	{
		return DataCubeExtraValue.getNew(getDbControl(), this, valueType, coordinateType);
	}

	private PositionBatcher positionBatcher;
	
	PositionBatcher getPositionBatcher()
		throws BaseException
	{
		if (positionBatcher == null)
		{
			positionBatcher = new PositionBatcher(getDbControl(), this);
		}
		return positionBatcher;
	}

	private MappingBatcher mappingBatcher;
	
	MappingBatcher getMappingBatcher()
		throws BaseException
	{
		if (mappingBatcher == null)
		{
			mappingBatcher = new MappingBatcher(getDbControl(), this);
		}
		return mappingBatcher;
	}

	/**
		Count the number of raw data mappings for the spot with the most
		mappings.
		@since 2.2
	*/
	int countSpotMappings()
	{
		if (mappingBatcher == null || mappingBatcher.getTotalInsertCount() == 0)
		{
			// No mappings to raw data has been inserted. Return 0 since the mapping
			// may not exists. See ticket #475
			return 0;
		}
		
		DynamicQuery query = new RawMappingQuery(this);
		Select count = Selects.expression(Aggregations.count(null, false), "mapCount");
		Expression selectedCount = Expressions.selected(count);
		query.select(count);
		query.group(Dynamic.column(VirtualColumn.COLUMN));
		query.group(Dynamic.column(VirtualColumn.POSITION));
		query.order(Orders.desc(selectedCount));
		query.setMaxResults(1);
		/*
			SELECT COUNT(*) AS mapCount 
			FROM Dymamic#RawParents raw
			WHERE raw.cube = ?
			GROUP BY raw.column, raw.position
			ORDER BY mapCount DESC
			LIMIT 1
		*/
		DynamicResultIterator result = query.iterate(getDbControl());
		try
		{
			int maxCount = 0;
			if (result.hasNext())
			{
				SqlResult r = result.next();
				maxCount = r.getInt(result.getIndex("mapCount"));
			}
			return maxCount;
		}
		catch (SQLException ex)
		{
			throw new BaseException(ex);
		}
		finally
		{
			result.close();
		}
	}

	/**
		Select data from Dynamic#RawParents table. Used for counting
		number of mappings between spots and raw data.
	
		@version 2.2
	*/
	private static final class RawMappingQuery
		extends DynamicQuery
	{
		
		private RawMappingQuery(DataCube dataCube)
		{
			super(dataCube, VirtualTable.RAWPARENTS, null);
		}
		
		/*
			From the AbstractSqlQuery class
			-------------------------------------------
		*/
		Select[] getDefaultSelects()
		{
			return null;
		}
		// -------------------------------------------
		/*
			From the DynamicQuery class
			-------------------------------------------
		*/
		/**
			No JEP functions can be used with this type of query.
			@since 2.15
		*/
		@Override
		public JepFunction[] getJepFunctions(DbControl dc, boolean restrictions)
		{
			return null;
		}
		// -------------------------------------------
	}

}
