/**
	$Id: LoggingInterceptor.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;


import net.sf.basedb.core.LogControl;

import org.hibernate.EmptyInterceptor;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.type.Type;

/**
	Interceptor that attaches to Hibernate sessions to provide
	logging of items that are updated/created/deleted. Do not attach
	this interceptor to a {@link SessionFactory} since this class is
	not thread-safe.
	<p>
	The interceptor uses <code>onSave</code>, <code>onDelete</code> and
	<code>onFlushDirty</code> to collect information about what has happened.
	The {@link LogManagerFactory} is asked to create {@link EntityLogger}:s
	for each modified/created/deleted item. The actual logging takes place
	in the <code>postFlush</code> method that calls {@link 
	EntityLogger#logChanges(LogManager, EntityDetails)} on each entity logger.
	<p>
	NOTE! For new items, the ID is not available until after the flush
	to the database has been done. Eg. new items has a 0 id when 
	{@link LogManagerFactory#getEntityLogger(LogManager, Object)} is called.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class LoggingInterceptor
	extends EmptyInterceptor
{

	private static final long serialVersionUID = -9108587050205509572L;
	private final LogControl logControl;
	private final LogManagerFactory logManagerFactory;
	private LogManager logManager;
	private List<EntityDetails> changes;
	
	/**
		Creates a new interceptor.
		@param logControl The LogControl that is managing the current transaction
		@param logManagerFactory The log manager factory to use for logging
	*/
	public LoggingInterceptor(LogControl logControl, LogManagerFactory logManagerFactory)
	{
		this.logControl = logControl;
		this.logManagerFactory = logManagerFactory;
	}
	
	/*
		From the Interceptor interface
		--------------------------------
	*/
	/**
		Adds a CREATE event to the log.
		@return Always false, since the state is not modified
	*/
	@Override
	public boolean onSave(Object entity, Serializable id, Object[] state,
			String[] propertyNames, Type[] types)
	{
		addChange(ChangeType.CREATE, entity, state, null, propertyNames, types);
		return false;
	}
	
	/**
		Adds an UPDATE event to the log.
		@return Always false, since the state is not changed
	*/
	@Override
	public boolean onFlushDirty(Object entity, Serializable id,
			Object[] currentState, Object[] previousState, String[] propertyNames,
			Type[] types)
	{
		addChange(ChangeType.UPDATE, entity, currentState, previousState, propertyNames, types);
		return false;
	}
	
	/**
		Adds a DELETE event to the log.
	*/
	@Override
	public void onDelete(Object entity, Serializable id, Object[] state,
			String[] propertyNames, Type[] types)
	{
		addChange(ChangeType.DELETE, entity, state, state, propertyNames, types);
	}

	/**
		Collects all log events and sends them to the each {@link EntityLogger}.
	*/
	@Override
	public void postFlush(Iterator entities)
	{
		if (changes != null && changes.size() > 0)
		{
			for (EntityDetails entry : changes)
			{
				entry.getEntityLogger().logChanges(logManager, entry);
			}
			// NOTE! A flush may happen more than once during the lifetime of a
			// transaction, so we must clear the list of changes
			changes.clear();
		}
	}

	/**
		Signals successful commit or rollback to the log manager.
		Note that this happens after the transaction/connection has been
		closed and no communication with the database can take place
		here.
	*/
	@Override
	public void afterTransactionCompletion(Transaction tx)
	{
		if (logManager != null)
		{
			if (tx.wasRolledBack())
			{
				logManager.afterRollback();
			}
			else
			{
				logManager.afterCommit();
			}
		}
		logManager = null;
		changes = null;
	}
	// --------------------------------------------

	/**
		If the log manager factory creates an entity logger, add
		information about the change to the internal list. The list
		is processed in the {@link #postFlush(Iterator)} method.
	 */
	private void addChange(ChangeType action, Object entity, Object[] state, 
			Object[] previousState, String[] propertyNames, Type[] types)
	{
		if (logManager == null)
		{
			this.logManager = logManagerFactory.getLogManager(logControl);
			this.changes = new LinkedList<EntityDetails>();
		}
		EntityLogger logger = logManagerFactory.getEntityLogger(logManager, entity);
		if (logger != null)
		{
			changes.add(new EntityDetails(logger, entity, action, 
					state, previousState, propertyNames, types));
		}

	}
	
	
	// This class is not serializable
	private void readObject(ObjectInputStream ois)
		throws ClassNotFoundException, IOException 
	{
		throw new NotSerializableException();
	}
	private void writeObject(ObjectOutputStream ois)
		throws IOException 
	{
		throw new NotSerializableException();
	}

}
