/*
	$Id: RlikeRestriction.java 5320 2010-04-21 09:48:29Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import java.util.Arrays;
import java.util.Collection;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.HibernateUtil;
import net.sf.basedb.util.EqualsHelper;

/**
	Checks if an expression matches a regular expression: 
	e RLIKE regexp (MySQL) or e ~ regexp (Postgres).

	@author Nicklas
	@version 2.8
	@see Restrictions#rlike(Expression, Expression)
	@base.modified $Date: 2010-04-21 11:48:29 +0200 (Wed, 21 Apr 2010) $
*/
public class RlikeRestriction
	implements Restriction 
{
	private final Expression e;
	private final Expression regexp;

	RlikeRestriction(Expression e, Expression regexp)
	{
		assert e != null : "e == null";
		assert regexp != null : "regexp == null";
		this.e = e;
		this.regexp = regexp;
	}

	/*
		From the QueryElement interface
		-------------------------------------------
	*/
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		if (query.getQueryType() == QueryType.HQL)
		{
			return "true = rlike(" + e.toQl(query, dc) + ", " + regexp.toQl(query, dc) + ")";
		}
		else
		{
			return HibernateUtil.getDbEngine().rlike(e.toQl(query, dc), regexp.toQl(query, dc));
		}
	}
	/**
		@return The expression
	*/
	public Collection<? extends QueryElement> getChildren()
	{
		return Arrays.asList(e, regexp);
	}
	// -------------------------------------------
	/*
		From the Object class
		-------------------------------------------
	*/
	@Override
	public String toString()
	{
		return e + " ~ " + regexp;
	}
	/**
		This restriction is equal to another RlikeRestriction if their 
		child restrictions are equal.
	*/
	@Override
	public boolean equals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		RlikeRestriction o = (RlikeRestriction)other;
		return EqualsHelper.equals(e, o.e) && EqualsHelper.equals(regexp, o.regexp);
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(e, regexp);
	}
	// -------------------------------------------
}

