/**
	$Id: MetadataWriter.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.Writer;
import java.nio.charset.Charset;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.File;
import net.sf.basedb.util.encode.EncoderDecoder;
import net.sf.basedb.util.encode.TabCrLfEncoderDecoder;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.util.formatter.ToStringFormatter;

/**
	Writer implementation for writing BFS metadata files. The start
	of file marker (BFSformat) is written immediately at object 
	construction. Use the {@link #bfsPrintSection(String, boolean, boolean)}
	to start a new section, and {@link #bfsPrintValue(String, String...)}
	to print a section entry. It is not recommended that any of
	the superclass methods (eg. {@link PrintWriter} methods} are used
	directly as it can create an invalid BFS metadata file.
	<p>
	
	This class can be subclassed to get specific functionality in specific
	use cases.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-08-13 10:50:27 +0200 (Fri, 13 Aug 2010) $
*/
public class MetadataWriter
	extends PrintWriter
{

	public static final String BOF_MARKER = "BFSformat";

	/**
		Utility method for creating a metadata writer when you have an
		output stream.
		@param out The output stream the metadata writer should print to
		@param filename Optional, the name of the file the output stream
			is printing to
	*/
	public static MetadataWriter create(OutputStream out, String filename)
	{
		if (out == null) throw new NullPointerException("out");
		MetadataWriter writer = new MetadataWriter(
				new OutputStreamWriter(out, Charset.forName("UTF-8")));
		writer.setFilename(filename);
		return writer;
	}
	
	/**
		Utility method for creating a metadata writer to a file in the BASE
		file system. The character set and MIME type on the file are automatically 
		updated.
		@param file The file in the BASE file system
	*/
	public static MetadataWriter create(File file)
	{
		if (file == null) throw new NullPointerException("file");
		file.setCharacterSet("UTF-8");
		file.setMimeType("text/plain");
		return create(file.getUploadStream(false), file.getName());
	}

	/**
		Utility method for creating a metadata writer to a file in the native
		file system. If the file doesn't exists, it is created.
		@param file The file in the native file system
	*/
	public static MetadataWriter create(java.io.File file)
		throws IOException
	{
		if (file == null) throw new NullPointerException("file");
		if (!file.createNewFile() && !file.exists()) 
		{
			throw new FileNotFoundException(file.toString());
		}
		return create(new FileOutputStream(file), file.getName());
	}

	
	private final EncoderDecoder encoder;
	private String filename;
	private String subtype;
	private Set<String> usedSections;
	private Set<String> usedKeys;
	private String currentSection;
	private boolean uniqueKeys;
	private int lineCount;
	private int sectionCount;
	private int valueCount;
	
	/**
		Create a new BFS metadata writer.
		
		@param out The parent writer which this writer will print to
	*/
	public MetadataWriter(Writer out)
	{
		super(out);
		this.usedSections = new HashSet<String>();
		this.usedKeys = new HashSet<String>();
		this.encoder = new TabCrLfEncoderDecoder(true);
	}
	
	/**
		Get the file name that this writer is printing to.
		@return The file name or null if not known
	*/
	public String getFilename()
	{
		return filename;
	}
	
	/**
		Set the file name that this writer is printing to.
	*/
	public void setFilename(String filename)
	{
		this.filename = filename;
	}
	
	/**
		Get the BFS subtype.
		@return The unencoded subtype, or null if no subtype was specified
			when the object was created
	*/
	public String getSubtype()
	{
		return subtype;
	}

	/**
		Set the BFS subtype for this writer. This method must be called
		before any printing method is called.
		@param subtype The subtype
		@throws IllegalStateException If data has already been written
	*/
	public void setSubtype(String subtype)
	{
		if (getLineCount() > 0)
		{
			throw new IllegalStateException("Can't change the subtype after data has been printed");
		}
		this.subtype = subtype;
	}
	
	/**
		Get the name of the current section.
		@return The unencoded name of the section, or null if no
			section has been started yet
	*/
	public String getCurrentSection()
	{
		return currentSection;
	}

	/**
		Checks if the current section requires entry names to be unique
		or not.
	*/
	public boolean getUniqueKeys()
	{
		return uniqueKeys;
	}
	
	/**
		Checks if the current section already has an entry for the given
		key or not.
		@param key The key to check
		@return TRUE if the current section already has an entry for this key,
			FALSE otherwise
	*/
	public boolean isUsedKey(String key)
	{
		return usedKeys.contains(key);
	}
	
	/**
		Encode a value for use in a BFS file. There is usually no need
		to call this method, since values are automatically encoded.
		@param value The value to encode, may be null
		@return The encoded value
	*/
	public String encodeValue(String value)
	{
		return encoder.encode(value);
	}

	/**
		Get the number of lines that has been written to this
		file so far, including all comment lines, empty lines,
		sections headers, etc.
	*/
	public int getLineCount()
	{
		return lineCount;
	}
	
	/**
		Get the number of sections that has been written to this
		file so far.
	*/
	public int getSectionCount()
	{
		return sectionCount;
	}
	
	/**
		Get the number of values that has been written to the current
		section so far.
	*/
	public int getValueCount()
	{
		return valueCount;
	}
	
	/**
		Print one or more empty lines to the metadata file.
		@param num The number of empty lines to print. If 0 or negative
			nothing is written
	*/
	public void bfsPrintEmptyLine(int num)
	{
		if (getLineCount() == 0) internalPrintBofMarker();
		for (int i = 0; i < num; ++i)
		{
			print("\n");
		}
		lineCount += num;
	}
	
	/**
		Prints one or more comment lines to the metadata file. A null
		comment is ignored. If the comment contains line-breaks and/or
		carriage returns, it will be split and written to multiple lines.
		@param comment The comment string
	*/
	public void bfsPrintComment(String comment)
	{
		if (comment == null) return;
		if (getLineCount() == 0) internalPrintBofMarker();
		String[] lines = comment.split("\\n|\\r|\\r\\n");
		for (String line : lines)
		{
			internalPrintComment(line);
		}
	}
	
	/**
		Start a new section.

		@param name The name of the section, null is not allowed
		@param unique TRUE if this section must be unique within the metadata
			file
		@param uniqueKeys TRUE if all keys in this section must be unique
		@throws NullPointerException If the section name is null
		@throws IllegalArgumentException If the name is not unique and
			the writer requires unique sections
	*/
	public void bfsPrintSection(String name, boolean unique, boolean uniqueKeys)
	{
		if (getLineCount() == 0) internalPrintBofMarker();
		if (name == null) throw new NullPointerException("name");
		if (unique)
		{
			if (usedSections.contains(name))
			{
				throw new IllegalArgumentException("Section '" + name + "' is not unique");
			}
		}
		usedSections.add(name);
		
		this.currentSection = name;
		this.uniqueKeys = uniqueKeys;
		usedKeys.clear();
		internalPrintSection(encodeValue(name));
	}
	
	/**
		Prints a section entry. A section must have been started with
		{@link #bfsPrintSection(String, boolean, boolean)} before this
		method is invoked.
		
		@param key The entry key; null is not allowed and it may not
			start with '[' or '#'
		@param values An array of string values, each value will be encoded
			independently and separated with a tab
		@throws NullPointerException If the key is null
		@throws IllegalArgumentException If the key starts with '[' or
			'#', or if the key is not unique within a section that requires
			unique keys
		@throws IllegalStateException If no section has been started
	*/
	public void bfsPrintValue(String key, String... values)
	{
		bfsPrintValue(key, null, values);
	}

	/**
		Prints a section entry. A section must have been started with
		{@link #bfsPrintSection(String, boolean, boolean)} before this
		method is invoked.
		
		@param key The entry key; null is not allowed and it may not
			start with '[' or '#'
		@param values An array of objects, each value will be be converted to
			a string by calling the toString() method
		@throws NullPointerException If the key is null
		@throws IllegalArgumentException If the key starts with '[' or
			'#', or if the key is not unique within a section that requires
			unique keys
		@throws IllegalStateException If no section has been started
	*/
	public <T> void bfsPrintValue(String key, T... values)
	{
		bfsPrintValue(key, null, values);
	}
	
	/**
		Prints a section entry. A section must have been started with
		{@link #bfsPrintSection(String, boolean, boolean)} before this
		method is invoked.
		
		@param key The entry key; null is not allowed and it may not
			start with '[' or '#'
		@param formatter A formatter that should be used to convert the
			values to strings, or null to simply use the toString() method
		@param values An array of objects, each value will be encoded
			independently and separated with a tab
		@throws NullPointerException If the key is null
		@throws IllegalArgumentException If the key starts with '[' or
			'#', or if the key is not unique within a section that requires
			unique keys
		@throws IllegalStateException If no section has been started
	*/
	public <T> void bfsPrintValue(String key, Formatter<T> formatter, T... values)
	{
		if (sectionCount == 0) throw new IllegalStateException("No section has been started");
		if (key == null) throw new NullPointerException("key");
		if (key.startsWith("[") || key.startsWith("#"))
		{
			throw new IllegalArgumentException("A key may not start with [ or #: '" + key + "'");
		}
		if (getUniqueKeys())
		{
			if (isUsedKey(key))
			{
				throw new IllegalArgumentException("Key '" + key + 
					"' is not unique in section '" + getCurrentSection() + "'");
			}
		}
		usedKeys.add(key);
		
		if (formatter == null) formatter = new ToStringFormatter<T>();
		StringBuilder value = new StringBuilder();
		if (values != null)
		{
			String sep = "";
			for (T v : values)
			{
				value.append(sep);
				value.append(encodeValue(formatter.format(v)));
				sep = "\t";
			}
		}
		internalPrintValue(encodeValue(key), value.toString());
	}
	
	/**
		Prints the beginning-of-file marker. Eg.
		<code>BFSformat&lt;tab&gt;subtype</code>
	*/
	protected void internalPrintBofMarker()
	{
		print(BOF_MARKER);
		String subtype = getSubtype();
		if (subtype != null) print("\t" + encodeValue(subtype));
		print("\n");
		lineCount++;
	}
	
	/**
		Print a single-line comment. This method should not be
		called with comment string that is null or contain
		line-breaks. Updates the line counter.
	*/
	protected void internalPrintComment(String comment)
	{
		print("#" + comment + "\n");
		lineCount++;
	}
	
	/**
		Prints a section header assuming that the name has already
		been encoded. Updates line and section counters and resets
		the value counter.
	*/
	protected void internalPrintSection(String encodedName)
	{
		print("[" + encodedName + "]\n");
		lineCount++;
		sectionCount++;
		valueCount = 0;
	}
	
	/**
		Prints a section entry assuming that the key and value has
		already been encoded. Updates the line and value counters.
	*/
	protected void internalPrintValue(String encodedKey, String encodedValue)
	{
		print(encodedKey + "\t" + encodedValue + "\n");
		lineCount++;
		valueCount++;
	}

}
