/**
	$Id: SequenceFilenameGenerator.java 5212 2010-01-12 13:21:49Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.bfs;

/**
	File name generator implementation that generates file names using
	a simple numerical counter. It is possible to give a prexix and/or
	suffix. For example, with prefix="file-" and suffix=".txt", the following
	file sequence is generated: file-1.txt, file-2.txt, file-3.txt
	This generator completely ignores the owner and suggested file names.
	
	@author Nicklas
	@version 2.15
	@base.modified $Date: 2010-01-12 14:21:49 +0100 (Tue, 12 Jan 2010) $
*/
public class SequenceFilenameGenerator
	implements FilenameGenerator<Object>
{
	private String prefix;
	private String suffix;
	private int nextFileNum;

	/**
		Create a new sequence file name generator. No prefix or suffix are
		used and the sequence starts at 1.
	*/
	public SequenceFilenameGenerator()
	{
		this.nextFileNum = 1;
	}
	
	/**
		Create a new sequence file name generator.
		@param prefix The prefix to use
		@param suffix The suffix to use
		@param firstFileNum The sequence number of the first file
	*/
	public SequenceFilenameGenerator(String prefix, String suffix, int firstFileNum)
	{
		this.prefix = prefix;
		this.suffix = suffix;
		this.nextFileNum = firstFileNum;
	}
	
	/*
		From the FilenameGenerator interface
		------------------------------------
	*/
	@Override
	public String getNextFilename(Object owner, String suggestedFilename)
	{
		StringBuilder sb = new StringBuilder();
		String prefix = getPrefix();
		String suffix = getSuffix();
		if (prefix != null) sb.append(prefix);
		sb.append(nextFileNum);
		nextFileNum++;
		if (suffix != null) sb.append(suffix);
		return sb.toString();
	}
	// -----------------------------------
	
	/**
		Get the prefix that is used for file name	generation. 
		@return The prefix or null if no prefix was given
	*/
	public String getPrefix()
	{
		return prefix;
	}
	
	/**
		Set the prefix that is used for file name generation.
		@param prefix The prefix or null to not use a prefix
	*/
	public void setPrefix(String prefix)
	{
		this.prefix = prefix;
	}
	
	/**
		Get the suffix that is used for file name	generation. 
		@return The suffix or null if no suffix was given
	*/
	public String getSuffix()
	{
		return suffix;
	}
	
	/**
		Set the suffix that is used for file name generation.
		@param suffix The suffix or null to not use a suffix
	*/
	public void setSuffix(String suffix)
	{
		this.suffix = suffix;
	}

	/**
		Get the next file number. This counter is increased for 
		each file that is generated (eg. each time {@link #getNextFilename(Object, String)}
		is called.
	*/
	public int getNextFileNumber()
	{
		return nextFileNum;
	}

	
}
