/**
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ExtraValue.CoordinateType;


/**
	Default field converter implementation that works mainly by
	checking names and/or external id values of formulas, extra
	values, experimental factors etc.
	
	@author nicklas
	@since 2.15
*/
public class StandardFieldConverter
	extends AbstractFieldConverter
{

	/**
		Create a new standard field converter.
		@param dc A DbControl to use for database access
		@param source The source bioassay set
	*/
	public StandardFieldConverter(DbControl dc, BioAssaySet source)
	{
		super(dc, source);
	}
	
	/**
		Find an assay field. This method will try to locate a field 
		in the following order:
		
		<ol>
		<li>Check if there is an experimental factor with the given name 
			(case-sensitivity depends on the database that is used).
		<li>Check if there is an assay-level annotation type with the
			given name.
		<li>If name is 'name', 'description' or another static property 
			(case-insensitive) the corresponding assay value is used.
		</ol>
		
		@param name The name to find
		@param required If TRUE and no field is found this method will
			throw an ItemNotFoundException
		@return The field or null
	*/
	@Override
	public AssayField getAssayField(String name, boolean required)
	{
		AssayField fld = null;
		if (fld == null)
		{
			fld = getExperimentalFactorField(name, name);
		}
		if (fld == null)
		{
			fld = getAssayAnnotationField(name, name);
		}
		if (fld == null)
		{
			fld = getAssayPropertyField(name, name);
		}
		if (fld == null && required)
		{
			throw new ItemNotFoundException("Unknown assay field: " + name +
				" (you can add this field as an experimental factor or assay annotation)");
		}
		return fld;
	}

	/**
		Find a reporter dynamic field. This method will try to locate a field 
		in the following order:
		
		<ol>
		<li>Check if there is an position extra value with the given name 
			(case-sensitivity depends on the database that is used).
		<li>Check if there is a formula with the given name 
			(case-sensitivity depends on the database that is used).
		<li>If name is 'name', 'externalId', 'description' or another static
			property (case-insensitive) the corresponding reporter value 
			is used.
		<li>Special case for "External ID" and "Internal ID" to simplify pass-throguh
			to the BFS importer which expects at least one of those columns.
		</ol>
		@param name The name to find
		@param required If TRUE and no field is found this method will
			throw an ItemNotFoundException
		@return The field or null
	*/
	@Override
	public DynamicField getReporterField(String name, boolean required)
	{
		DynamicField fld = getExtraValueField(name, CoordinateType.POSITION, name);
		if (fld == null)
		{
			fld = getFormulaField(name, name);
		}
		if (fld == null)
		{
			fld = getReporterProperty(name, name);
		}
		if (fld == null && required)
		{
			throw new ItemNotFoundException("Unknown reporter field: " + name +
				" (you can add this field as a formula or extra value)");
		}
		return fld;
	}

	/**
		Find a spot dynamic field. This method will try to locate a field 
		in the following order:
		
		<ol>
		<li>Check if there is a spot extra value with the given name 
			(case-sensitivity depends on the database that is used).
		<li>Check if there is a formula with the given name 
			(case-sensitivity depends on the database that is used).
		<li>Check if there is a raw data property that matches the
			name (case-insensitive)
		<li>Special values: Ch 1, Ch2, etc. to get channel intensities 
			(case-insensitive)
		</ol>
		@param name The name to find
		@param required If TRUE and no field is found this method will
			throw an ItemNotFoundException
		@return The field or null
	*/
	@Override
	public DynamicField getSpotField(String name, boolean required)
	{
		DynamicField fld = getExtraValueField(name, CoordinateType.SPOT, name);
		if (fld == null)
		{
			fld = getFormulaField(name, name);
		}
		if (fld == null)
		{
			fld = getRawDataProperty(name, name);
		}
		if (fld == null)
		{
			fld = getIntensityField(name, name);
		}
		if (fld == null && required)
		{
			throw new ItemNotFoundException("Unknown spot field: " + name +
				" (you can add this field as a formula or extra value)");
		}		
		return fld;
	}

}
