/**
	$Id $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.formatter;


/**
 	Formats a value with optional prefix and suffix. This formatter uses
 	a parent formatter to format the actual value and then adds
 	prefix and suffix as needed to the start and end of the
 	formatted value.
 	<p>
 	The {@link #parseString(String)} method will first check if
 	the string start or ends with the given prefix/suffix. If it does, 
 	the prefix/suffix is removed before parsing is delegated to the 
 	parent.
 	
	@author nicklas
	@version 2.9
	@base.modified $Date: 2008-09-25 09:01:30 +0200 (Thu, 25 Sep 2008) $
*/
public class PrefixSuffixFormatter<T>
	implements Formatter<T>
{
	
	/**
		The character representing a non-breaking space. Usefule to
		avoid line wrapping between prefix/suffix and value.
	*/
	public static final char NBSP = 0xA0;

	private final String prefix;
	private final Formatter<T> parent;
	private final String suffix;
	
	/**
		Create a new formatter.
		@param prefix The prefix, or null to not use any prefix
		@param parent The parent formatter, if null a {@link ToStringFormatter}
			is automatically created
		@param suffix The suffix, or null to not use any suffix
	*/
	public PrefixSuffixFormatter(String prefix, Formatter<T> parent, String suffix)
	{
		this.prefix = prefix == null ? "" : prefix;
		this.parent = parent == null ? new ToStringFormatter<T>() : parent;
		this.suffix = suffix == null ? "" : suffix;
	}
	
	/*
		From the Formatter interface
		-------------------------------------------
	*/
	public String format(T value)
	{
		return prefix + parent.format(value) + suffix;
	}
	public T parseString(String value)
	{
		if (value != null)
		{
			if (prefix.length() > 0 && value.startsWith(prefix))
			{
				value = value.substring(prefix.length());
			}
			if (suffix.length() > 0 && value.endsWith(suffix))
			{
				value = value.substring(0, value.length()-suffix.length());
			}
		}
		return parent.parseString(value);
	}
	// -------------------------------------------

}
