/**
	$Id: ConverterFactory.java 4515 2008-09-11 20:08:14Z jari $

	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jep.convert;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.ItemAlreadyExistsException;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Restriction;

import org.nfunk.jep.ASTConstant;
import org.nfunk.jep.ASTFunNode;
import org.nfunk.jep.ASTVarNode;
import org.nfunk.jep.function.Abs;
import org.nfunk.jep.function.Add;
import org.nfunk.jep.function.Divide;
import org.nfunk.jep.function.Exp;
import org.nfunk.jep.function.If;
import org.nfunk.jep.function.Logarithm;
import org.nfunk.jep.function.Multiply;
import org.nfunk.jep.function.NaturalLogarithm;
import org.nfunk.jep.function.Not;
import org.nfunk.jep.function.SquareRoot;
import org.nfunk.jep.function.Subtract;
import org.nfunk.jep.function.UMinus;

/**
	Keep {@link JepConversionFunction} objects for converting built-in
	JEP object to Query API {@link Expression} or {@link Restriction}
	objects.

	@author nicklas
	@version 2.4
	@see JepConversionFunction
	@base.modified $Date: 2008-09-11 22:08:14 +0200 (Thu, 11 Sep 2008) $
*/
public class ConverterFactory
{

	private static Map<Object, JepConversionFunction<?>> converters = 
		new HashMap<Object, JepConversionFunction<?>>();
	
	/**
		Register the built-in functions, operators and node types.
	*/
	static
	{
		// Basic nodes types
		converters.put(ASTFunNode.class, new ASTFunNodeConverter());
		converters.put(ASTVarNode.class, new ASTVarNodeConverter());
		converters.put(ASTConstant.class, new ASTConstantConverter());
		
		// Built-in operators
		converters.put(Add.class, new AddConverter());
		converters.put(Subtract.class, new SubtractConverter());
		converters.put(Multiply.class, new MultiplyConverter());
		converters.put(Divide.class, new DivideConverter());
		converters.put(UMinus.class, new UMinusConverter());
		converters.put("==", new EqualsConverter());
		converters.put("!=", new NotEqualsConverter());
		converters.put(">", new GreaterConverter());
		converters.put(">=", new GreaterEqualsConverter());
		converters.put("<", new LessConverter());
		converters.put("<=", new LessEqualsConverter());
		converters.put("&&", new AndConverter());
		converters.put("||", new OrConverter());
		converters.put(Not.class, new NotConverter());
		
		// Built-in functions
		converters.put(Logarithm.class, new LogarithmConverter());
		converters.put(NaturalLogarithm.class, new NaturalLogarithmConverter());
		converters.put(SquareRoot.class, new SquareRootConverter());
		converters.put(Abs.class, new AbsConverter());
		converters.put(Exp.class, new ExpConverter());
		converters.put(If.class, new IfConverter());
		
		// Constant values
		converters.put(Integer.class, new IntegerConverter());
		FloatConverter fc = new FloatConverter();
		converters.put(Float.class, fc);
		converters.put(Double.class, fc);
		converters.put(String.class, new StringConverter());
			
	}
	
	/**
		Get a converter for the specified built-in JEP object.
		@param o The object to convert
		@return A {@link JepConversionFunction} or null if no converter is found
	*/
	public static JepConversionFunction<?> getConverter(Object o)
	{
		if (o == null) return null;
		return converters.get(o);
	}
	
	/**
		Get a converter for the specified built-in JEP object by checking if
		a converter has been registered for it's class.
		@param o The object to convert
		@return A {@link JepConversionFunction} or null if no converter is found
	*/
	public static JepConversionFunction<?> getConverterByClass(Object o)
	{
		if (o == null) return null;
		return converters.get(o.getClass());
	}
	
	/**
		Add a converter for a specific object. It can for example, be the class
		of a JEP function, the symbol for an operator:
		<pre class="code">
registerConverter(Add.class, new AddConverter());
registerConverter("==", new EqualsConverter());
</pre>
		
		@param o The object defining what the converter can convert
		@param converter The converter
		@throws ItemAlreadyExistsException If a converter already exists for the given class
	*/
	public static void registerConverter(Object o, JepConversionFunction<?> converter)
	{
		if (o == null) throw new NullPointerException("o");
		if (converter == null) throw new NullPointerException("converter");
		if (converters.containsKey(o))
		{
			throw new ItemAlreadyExistsException("A converter for is already registered for object: " + o);
		}
		synchronized (converters)
		{
			converters.put(o, converter);
		}
	}
	
	/**
		Get all registered converters.
	*/
	public static Collection<JepConversionFunction<?>> getConverters()
	{
		return converters.values();
	}
	
}
