/**
	$Id: JobAgentInfo.java 5590 2011-03-16 11:48:47Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.jobagent;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.util.Values;

/**
	This class can hold information about the status of a job agent.

	@author nicklas
	@version 2.0
	@see JobAgentConnection#getInfo(boolean)
	@base.modified $Date: 2011-03-16 12:48:47 +0100 (Wed, 16 Mar 2011) $
*/
public class JobAgentInfo
{

	private final long created;
	private final Boolean paused;
	private final Integer cpuUsage;
	private final Long usedMemory;
	private final Long totalMemory;
	private final Set<JobInfo> jobs;
	
	/**
		Create a new info object and set all values to null.
	*/
	public JobAgentInfo()
	{
		this.created = System.currentTimeMillis();
		this.jobs = null;
		this.cpuUsage = null;
		this.totalMemory = null;
		this.usedMemory = null;
		this.paused = null;
	}
	
	/**
	 	Create a new info object.
	 	@param paused If the job agent is paused or running, or null if not know
		@param cpuUsage The CPU usage in percent, or null if not known
		@param usedMemory The used memory in bytes, or null if not known
		@param totalMemory The total memory in bytes, or null if not known
		@param jobs Information about jobs that are currently executing or null if not known
	 */
	public JobAgentInfo(Boolean paused, Integer cpuUsage, Long usedMemory, Long totalMemory, Set<JobInfo> jobs)
	{
		this.created = System.currentTimeMillis();
		this.paused = paused;
		this.jobs = jobs;
		this.cpuUsage = cpuUsage;
		this.usedMemory = usedMemory;
		this.totalMemory = totalMemory;
	}
	
	/**
		Create a new info object and parse the answer from a {@link JobAgentConnection}
		to get the values.
		@param answer The answer from <code>info</code> request to the job agent
	*/
	public JobAgentInfo(String answer)
	{
		Map<String, List<String>> headers = JobAgentConnection.parseAnswer(answer);
		this.created = System.currentTimeMillis();
		String tempPaused = getValue(headers, "Status");
		this.paused = tempPaused == null ? null : "Paused".equals(tempPaused);
		this.cpuUsage = Values.getInteger(getValue(headers, "Cpu"), null);
		this.totalMemory = Values.getLong(getValue(headers, "Total memory"), null);
		this.usedMemory = Values.getLong(getValue(headers, "Used memory"), null);
		this.jobs = new HashSet<JobInfo>();
		List<String> jj = getValues(headers, "Job");
		if (jj != null)
		{
			for (String id : jj)
			{
				Integer jobId = Values.getInteger(id, null);
				jobs.add(new JobInfo(jobId, headers));
			}
		}
	}
	
	/**
		Get the age in milliseconds of this object. The age can be used by
		client applications that caches the information to determine when it is
		time to issue a new request.
		@return The age in milliseconds
	*/
	public long getAge()
	{
		return System.currentTimeMillis() - created;
	}
	
	/**
		If the job agent is paused or running.
		@return TRUE if the job agent is pasued, FALSE if it is running, or null
			if not known
	*/
	public Boolean isPaused()
	{
		return paused;
	}
	
	/**
		Get the CPU usage in percent.
		@return The CPU usage or null if not known
	*/
	public Integer getCpuUsage()
	{
		return cpuUsage;
	}
	
	/**
		Get the total memory on the job agent's server in bytes.
		@return The total memory or null if not known
	*/
	public Long getTotalMemory()
	{
		return totalMemory;
	}
	
	/**
		Get the used memory on the job agent's server in bytes.
		@return The used memory or null if not known
	*/
	public Long getUsedMemory()
	{
		return usedMemory;
	}
	
	/**
		Get detailed information about the running jobs.
		@return A set with job information or null if not known
		@since 2.6
	*/
	public Set<JobInfo> getJobInfo()
	{
		return jobs;
	}
	
	/**
		Convert the information to a string that is suitable for sening as an
		answer to a {@link JobAgentConnection} object.
	*/
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		if (paused != null) {
			sb.append("Status:").append(paused ? "Paused\n" : "Running\n");
		}
		if (cpuUsage != null) sb.append("Cpu:").append(cpuUsage).append("\n");
		if (totalMemory != null) sb.append("Total memory:").append(totalMemory).append("\n");
		if (usedMemory != null) sb.append("Used memory:").append(usedMemory).append("\n");
		if (jobs != null)
		{
			for (JobInfo job : jobs)
			{
				sb.append(job.toString());
			}
		}
		return sb.toString();
	}
	
	private String getValue(Map<String, List<String>> headers, String header)
	{
		List<String> values = headers.get(header);
		String value = null;
		if (values != null && values.size() > 0)
		{
			value = values.get(0);
		}
		return value;
	}
	
	private List<String> getValues(Map<String, List<String>> headers, String header)
	{
		return headers.get(header);
	}
}
