/**
	$Id$

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DerivedBioAssay;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.NodeAttribute;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.Node.Type;
import net.sf.basedb.util.overview.node.ChildNodeDirection;

/**
	Validator implementation for derived bioassays. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_DERIVEDBIOASSAY}
	<li>Access denied: {@link Validator#DENIED_DERIVEDBIOASSAY}
	<li>Parent extract mismatch: {@link Validator#PARENT_EXTRACT_MISMATCH}
	<li>Incorrect related subtypes: {@link Validator#INCORRECT_DERIVEDBIOASSAY_TYPE},
		{@link Validator#INCORRECT_PHYSICALBIOASSAY_TYPE}, {@link Validator#INCORRECT_EXTRACT_TYPE}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-02-16 20:29:47 +0100 (må, 16 feb 2009) $
*/
public class DerivedBioAssayValidator
	extends NameableNodeValidator<DerivedBioAssay>
{
	
	public DerivedBioAssayValidator()
	{
		super(Validator.MISSING_DERIVEDBIOASSAY, Validator.DENIED_DERIVEDBIOASSAY);
	}

	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		If the parent raw bioassay has a link to an extract, we ignore the missing
		derived bioassay.
		@since 3.0
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		if (parentNode.hasAttribute(NodeAttribute.EXTRACT)) return false;
		return super.preMissingItem(dc, context, parentNode);
	}
	
	/**
		@return Suggested fix is to add a derived bioassay as parent to the raw bioassay.
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node rawBioAssayNode)
	{
		return new Fix("Add parent bioassay to raw bioassay", rawBioAssayNode.getItem());
	}
	
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode) 
	{
		super.postValidate(dc, context, node, parentNode);
		
		if (parentNode != null)
		{
			Node parentItemNode = parentNode.getNodeType() == Type.ITEM ? parentNode : parentNode.getParent();
			Item parentType = parentItemNode.getItemType();

			if (parentType == Item.PHYSICALBIOASSAY)
			{
				checkPhysicalBioAssaySubtype(dc, context, node, parentItemNode);
			}
			else if (parentType == Item.DERIVEDBIOASSAY)
			{
				if (node.getChildNodeDirection() == ChildNodeDirection.FORWARD)
				{
					checkDerivedBioAssaySubtype(dc, context, node, parentItemNode);
				}
				else
				{
					checkDerivedBioAssaySubtype(dc, context, parentItemNode, node);
				}
			}
		}
		
		// Check that parent items make sense
		DerivedBioAssay bioAssay = (DerivedBioAssay)node.getItem(dc);
		Extract extract = null;
		DerivedBioAssay parentBioAssay = null;
		Extract parentBioAssayExtract = null;
		PhysicalBioAssay physicalBioAssay = null;
		try
		{
			extract = bioAssay.getExtract();
			physicalBioAssay = bioAssay.getPhysicalBioAssay();
			if (!bioAssay.isRoot())
			{
				parentBioAssay = bioAssay.getParent();
				parentBioAssayExtract = parentBioAssay.getExtract();
			}
		}
		catch (PermissionDeniedException ex)
		{}
		
		// Check that extract is same on the child and parent bioassays
		if (parentBioAssayExtract != null)
		{
			if (!parentBioAssayExtract.equals(extract))
			{
				context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
					"The parent bioassay '" + parentBioAssay.getName() + "' has different extract: " + 
						parentBioAssayExtract.getName(), 
					new Fix("Change extract on bioassay", bioAssay),
					new Fix("Change extract on parent bioassay", parentBioAssay)
				);
			}
		}
		
		// Check that the extract is part of the physical bioassay
		if (extract != null && physicalBioAssay != null)
		{
			if (!physicalBioAssay.getCreationEvent().isSource(extract))
			{
				context.createFailure(Validator.PARENT_EXTRACT_MISMATCH, node, 
					"The extract '" + extract.getName() + "' is not found on the physical bioassay: " + 
						physicalBioAssay.getName(),
					new Fix("Change extract on bioassay", bioAssay),
					new Fix("Change extracts on physical bioassay", physicalBioAssay)
				);
			}
		}
	}

	
	// ----------------------------

	/**
		Check the subtype of the physical bioassay that is the parent to a derived bioassay. Given that
		the derived bioassay has a non-null subtype, which has a related PHYISICALBIOASSAY subtype, this
		method checks if the physical bioassay has the same subtype. If there is no related
		subtype for the derived bioassay, any subtype for the physical bioassay is accepted.
	*/
	public static void checkPhysicalBioAssaySubtype(DbControl dc, OverviewContext context, Node derivedBioAssayNode, Node physicalBioAssayNode)
	{
		DerivedBioAssay dba = (DerivedBioAssay)derivedBioAssayNode.getItem(dc);
		PhysicalBioAssay pba = (PhysicalBioAssay)physicalBioAssayNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, dba, Item.PHYSICALBIOASSAY, Item.DERIVEDBIOASSAY);
			ItemSubtype subtype = pba.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = dba.getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related physical bioassay subtype of '" + mySubtype.getName() + "' to '" + subtype.getName() + "'", mySubtype);
				if (expectedSubtype.getMainItemType() == Item.PHYSICALBIOASSAY)
				{
					context.createFailure(Validator.INCORRECT_PHYSICALBIOASSAY_TYPE, derivedBioAssayNode, 
							"Expected a physical bioassay with subtype: " + expectedSubtype.getName(),
							new Fix("Change subtype of '" + pba.getName() + "' to '" + expectedSubtype.getName() + "'", pba),
							new Fix("Change subtype of '" + derivedBioAssayNode.getTitle() + "'", dba),
							subtypeFix
						);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_PHYSICALBIOASSAY_TYPE, derivedBioAssayNode, 
							"Expected a parent bioassay with subtype: " + expectedSubtype.getName(),
							new Fix("Change subtype of '" + derivedBioAssayNode.getTitle() + "'", dba),
							subtypeFix
						);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	/**
		Check the subtype of the derived bioassay that is the parent to a child derived bioassay. Given that
		the child bioassay has a non-null subtype, which has a related DERIVEDBIOASSAY subtype, this
		method checks if the parent bioassay has the same subtype. If there is no related
		subtype for the child bioassay, any subtype for the parent bioassay is accepted.
	*/
	public static void checkDerivedBioAssaySubtype(DbControl dc, OverviewContext context, Node childNode, Node parentNode)
	{
		DerivedBioAssay child = (DerivedBioAssay)childNode.getItem(dc);
		DerivedBioAssay parent = (DerivedBioAssay)parentNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, child, Item.DERIVEDBIOASSAY, Item.PHYSICALBIOASSAY);
			ItemSubtype subtype = parent.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = child.getItemSubtype();
				Fix subtypeFix = subtype == null || subtype.equals(mySubtype) ? null : 
					new Fix("Change related derived bioassay subtype of '" + mySubtype.getName() + "' to '" + subtype.getName() + "'", mySubtype);
				if (expectedSubtype.getMainItemType() == Item.DERIVEDBIOASSAY)
				{
					context.createFailure(Validator.INCORRECT_DERIVEDBIOASSAY_TYPE, childNode, 
						"Expected a parent bioassay with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + parent.getName() + "' to '" + expectedSubtype.getName() + "'", parent),
						new Fix("Change subtype of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
				}
				else
				{
					context.createFailure(Validator.INCORRECT_DERIVEDBIOASSAY_TYPE, childNode, 
						"Expected a physical bioassay with subtype: " + expectedSubtype.getName(),
						new Fix("Change subtype of '" + childNode.getTitle() + "'", child),
						subtypeFix
					);
				}
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}

	
	/**
		Check the subtype of the extract that is the linked with a derived bioassay. Given that
		the bioassay has a non-null subtype, which has a related EXTRACT subtype, this
		method checks if the extract has the same subtype. If there is no related
		subtype for the bioassay, any subtype for the extract is accepted.
	*/
	public static void checkExtractSubtype(DbControl dc, OverviewContext context, Node bioAssayNode, Node extractNode)
	{
		DerivedBioAssay bioAssay = (DerivedBioAssay)bioAssayNode.getItem(dc);
		Extract extract = (Extract)extractNode.getItem(dc);
		
		try
		{
			ItemSubtype expectedSubtype = ItemSubtype.getRelatedSubtype(dc, bioAssay, Item.EXTRACT, 0);
			ItemSubtype subtype = extract.getItemSubtype();
			
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = bioAssay.getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related extract subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);
				context.createFailure(Validator.INCORRECT_EXTRACT_TYPE, bioAssayNode, 
					"Expected an extract with subtype: " + expectedSubtype.getName(), 
					new Fix("Change subtype of '" + extract.getName() + "' to '" + expectedSubtype.getName() + "'", extract),
					new Fix("Change subtype or extract of '" + bioAssayNode.getTitle() + "'", bioAssay),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
	}


}