import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.FlatFileParser.Line;

/*
 $Id$

 Copyright (C) 2005 Johan Enell, Nicklas Nordborg
 Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg
 Copyright (C) 2007 Johan Enell
 
 This file is part of BASE - BioArray Software Environment.
 Available at http://base.thep.lu.se/

 BASE is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either version 3
 of the License, or (at your option) any later version.

 BASE is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with BASE. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 
 @base.modified $Date$
 @author Enell
 @version 2.0
 */
public class TestFlatFileParser
{
	static boolean ok = true;

	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing flat file parser");

		FlatFileParser ffp = test_create("\"(.*)=(.*)\"", "\"?(.*?)\"?(\\t\"?(.*?)\"?){10,}", null, "\\t", null, null, 20, 10, 0);

		test_parse(ffp, "data/test.parse.hskg.plates.txt", "\\3\\", "\\cluster_id\\", "Row:\\1\\, Col:\\2\\");
		test_parse(ffp, "data/test.parse.genepix.result.gpr", "\\3\\", "\\Name\\", "B\\Block\\R\\Row\\C\\Column\\", "=col('Dia.')",
			"=col('Dia.') / 2", "=\\Name\\+':'+\\ID\\");

		ffp = test_create("(.*)=,(.*)", null, null, ",", null, "\\[(.*)\\]", 20, 10, 0);
		test_parse_section(ffp, "data/test.parse.printmap.mwbr");

		write("++Testing file parser " + (ok ? "OK" : "Failed") + "\n");
		return ok;
	}

	public static FlatFileParser test_create(String header, String data_header, String data_footer, String data, String ignore, String section, int max_unknown_lines, int min_data_columns, int max_data_columns)
	{
		FlatFileParser ffp = new FlatFileParser();
		if (header != null) ffp.setHeaderRegexp(Pattern.compile(header));
		if (data_header != null) ffp.setDataHeaderRegexp(Pattern.compile(data_header));
		if (data_footer != null) ffp.setDataFooterRegexp(Pattern.compile(data_footer));
		if (data != null) ffp.setDataSplitterRegexp(Pattern.compile(data));
		ffp.setTrimQuotes(true);
		if (ignore != null) ffp.setIgnoreRegexp(Pattern.compile(ignore));
		if (section != null) ffp.setSectionRegexp(Pattern.compile(section));
		ffp.setMaxUnknownLines(max_unknown_lines);
		ffp.setMinDataColumns(min_data_columns);
		ffp.setMaxDataColumns(max_data_columns);
		return ffp;
	}

	public static void test_parse(FlatFileParser ffp, String file, String... mappings)
	{
		try
		{
			InputStream in = FileUtil.getInputStream(new java.io.File(file));
			test_parse(ffp, file, in, mappings);
		}
		catch (Throwable ex)
		{
			write("--Parse FAILED (" + file + ")");
			ex.printStackTrace();
			ok = false;
		}
	}


	public static void test_parse(FlatFileParser ffp, String file, InputStream in, String... mappings)
	{
		try
		{
			ffp.setInputStream(in, "ISO-8859-1");
			FlatFileParser.LineType last_line_type = ffp.parseHeaders();
			int i = 0;
			int headers = ffp.getLineCount();
			while (i < headers)
			{
				write(i, ffp.getLine(i));
				i++;
			}
			if (!TestUtil.getSilent()) write("--Column headers: " + ffp.getColumnHeaders());
			if ((last_line_type == FlatFileParser.LineType.DATA_HEADER) || (last_line_type == FlatFileParser.LineType.DATA))
			{
				Mapper[] mappers = new Mapper[mappings.length];
				for (int j = 0; j < mappings.length; ++j)
				{
					mappers[j] = ffp.getMapper(mappings[j]);
				}
				while (ffp.hasMoreData() && i < 20 + headers)
				{
					write(i++, ffp.nextData(), mappers);
				}
			}
			in.close();
			write("--Parse OK (" + file + ")");
		}
		catch (Throwable ex)
		{
			write("--Parse FAILED (" + file + ")");
			ex.printStackTrace();
			ok = false;
		}
	}

	public static void test_parse_section(FlatFileParser ffp, String file)
	{
		try
		{
			Set<String> sections = new HashSet<String>(Arrays.asList("FileInformation", "Block1", "Block2", "Block3", "Block4", "mapping")); 
			InputStream in = FileUtil.getInputStream(new java.io.File(file));
			ffp.setInputStream(in, "ISO-8859-1");
			FlatFileParser.Line section;
			while (ffp.hasMoreSections())
			{
				section = ffp.nextSection();
				sections.remove(section.name());
				ffp.parseHeaders();
				ffp.nextData();
			}
			in.close();
			if (!sections.isEmpty())
			{
				throw new Exception("Didn't find all sections in (" + file + "), missed: " + sections);
			}
			write("--Parse sections OK (" + file + ")");
		}
		catch (Throwable ex)
		{
			write("--Parse sections FAILED (" + file + ")");
			ex.printStackTrace();
			ok = false;
		}
	}

	static void write(String message)
	{
		System.out.println(message);
	}

	static void write(int i, FlatFileParser.Line line)
	{
		if (!TestUtil.getSilent())
		{
			if (line.type() == FlatFileParser.LineType.HEADER)
			{
				write("Line " + i + ": \t" + line.type() + "\t{" + line.name() + "} -> {" + line.value() + "}");
			}
			else
			{
				String s = line.line();
				if (s.length() > 30) s = s.substring(0, 30);
				write("Line " + i + ": \t" + line.type() + "\t" + s);
			}
		}
	}

	static void write(int i, FlatFileParser.Data data, Mapper[] mappings)
	{
		if (!TestUtil.getSilent())
		{
			StringBuffer sb = new StringBuffer();
			sb.append("Line " + i + ": \tDATA(" + data.columns() + ")\t");
			int max = data.columns();
			if (max > 5) max = 5;
			for (int j = 0; j < max; j++)
			{
				sb.append(j + "={" + data.get(j) + "}\t");
			}
			if (max < data.columns()) sb.append((data.columns() - max) + " more...");
			if (mappings != null)
			{
				sb.append("\n");
				for (Mapper m : mappings)
				{
					sb.append(m+"={" + m.getValue(data) + "}\t");
				}
			}
			write(sb.toString());
		}
	}

	public static void testHeaderLine(Line line, String name, String value)
	{
		if (!line.type().equals(FlatFileParser.LineType.HEADER))
		{
			throw new BaseException("Line is no header: "+line);
		}
		if (!line.name().equals(name))
		{
			throw new BaseException("Expected header name '"+name+"' on line '"+line+"'");
		}
		if (!line.value().equals(value))
		{
			throw new BaseException("Expected header value '"+value+"' on line '"+line+"'");
		}
	}

	public static void testHeaderLine(Line line, String name, String splitter, String... expected)
	{
		testHeaderLine(line, name, splitter, Arrays.asList(expected));
	}
	
	public static void testHeaderLine(Line line, String name, String splitter, List<String> expected)
	{
		if (!line.type().equals(FlatFileParser.LineType.HEADER))
		{
			throw new BaseException("Line is no header: "+line);
		}
		if (!line.name().equals(name))
		{
			throw new BaseException("Expected header name '"+name+"' on line '"+line+"'");
		}
		String[] lineValues = line.value().split(splitter);
		for (int i = 0; i < lineValues.length; ++i)
		{
			String fnd = lineValues[i];
			String exp = expected.get(i);
			if (!fnd.equals(exp))
			{
				throw new BaseException("Expected header value '" + exp + "' at index " + i + " on line '" + line + "'" );
			}
		}
	}

	
	public static void testLine(Line line, String name)
	{
		if (!line.line().equals(name))
		{
			throw new BaseException("Line '"+line+"' doesn't match '"+name+"'");
		}
	}

	public static void testSection(Line line, String name)
	{
		if (!line.type().equals(FlatFileParser.LineType.SECTION))
		{
			throw new BaseException("Line is no section: "+line);
		}
		if (!line.name().equals(name))
		{
			throw new BaseException("Expected section name '"+name+"' on line '"+line+"'");
		}
	}
	
	public static void testDataSection(FlatFileParser ffp, List<Pattern[]> patterns, Line section)
		throws IOException
	{
		int index = 0;
		while (ffp.hasMoreData())
		{
			Data data = ffp.nextData();
			if (index < patterns.size()) 
			{
				Pattern[] p = patterns.get(index);
				if (p != null) testDataLine(data, p);
			}
			index++;
		}
		
		if (index != patterns.size())
		{
			throw new BaseException("Unexpected number of data lines. Found "+index+
					"; expected "+patterns.size() + " in section " + section);
		}
	}
	
	public static void testDataLine(Data data, Pattern[] regexp)
	{
		if (data.columns() != regexp.length)
		{
			throw new BaseException("Expected "+regexp.length+" columns on line "+data);
		}
		for (int i = 0; i < regexp.length; ++i)
		{
			Pattern p = regexp[i];
			if (p != null)
			{
				String value = data.get(i);
				Matcher m = p.matcher(value);
				if (!m.matches())
				{
					throw new BaseException("Value '"+value+"' at column " + i + " dosen't match the pattern '"+p+"' on line: " + data);
				}
			}
		}
	}
	
}
